/*
 * @(#)ChangeAttributeCommand.java 5.1
 *
 */
package CH.ifa.draw.standard;

import CH.ifa.draw.DrawPlugin;

import CH.ifa.draw.framework.DrawingEditor;
import CH.ifa.draw.framework.DrawingView;
import CH.ifa.draw.framework.Figure;
import CH.ifa.draw.framework.FigureEnumeration;

import CH.ifa.draw.util.Command;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JTextField;


/**
 * Command to change a named figure attribute.
 * <p>
 * The command's effects are undoable step by step.
 * Each use of the apply button can be undone separately.
 * So this command doesn't need to inherit UndoableCommand.
 * </p>
 */
public class QueryAttributeCommand extends Command {
    private JFrame lastParent = null;
    private JDialog dialog = null;
    String fAttribute;
    JTextField input = null;
    private Class type;
    String displayName;

    /**
     * Constructs a query attribute command.
     * @param displayName the dialog name
     * @param name the command name
     * @param attributeName the name of the attribute to be changed
     * @param type the attribute type (class)
     * @param editor the target editor
     * @param parent the parent Frame
     */
    public QueryAttributeCommand(String displayName, String name,
                                 String attributeName, Class type) {
        super(name);
        this.fAttribute = attributeName;
        this.displayName = displayName;
        this.type = type;
    }

    private JDialog getDialog() {
        // We can reuse an existing dialog, but we must create a new dialog
        // when the gui frame has been closed and recreated inbetween.
        // This is possible because the menu command objects are kept across
        // gui instances by the MenuManager
        // TODO: Redesign of the MenuManager or apply the same mechanism to
        //       all other editor dialogs (see bug #99).
        JFrame parent = DrawPlugin.getGui().getFrame();
        if (dialog != null) {
            if (lastParent == parent) {
                // reuse the existing dialog
                return dialog;
            } else {
                // or release the old one
                dialog.dispose();
            }
        }
        dialog = new JDialog(parent, "Change " + displayName + ":");
        lastParent = parent;

        input = new JTextField();
        input.addKeyListener(new KeyAdapter() {
                public void keyPressed(KeyEvent e) {
                    if (e.getKeyCode() == KeyEvent.VK_ENTER) {
                        apply(type);
                    } else if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
                        dialog.setVisible(false);
                    }
                }
            });
        dialog.getContentPane().add(input, BorderLayout.CENTER);

        JButton apply = new JButton(" Apply ");
        apply.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    apply(type);
                }
            });

        JButton update = new JButton(" Update ");
        update.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    execute();
                }
            });

        JButton ok = new JButton(" OK ");
        ok.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    apply(type);
                    dialog.setVisible(false);
                }
            });

        JButton cancel = new JButton(" Cancel ");
        cancel.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    resetFocus();
                    dialog.setVisible(false);
                }
            });

        JPanel panel = new JPanel();
        panel.setLayout(new FlowLayout());
        panel.add(apply);
        panel.add(update);
        panel.add(ok);
        panel.add(cancel);
        dialog.getContentPane().add(panel, BorderLayout.PAGE_END);
        dialog.pack();
        return dialog;
    }

    protected DrawingEditor getEditor() {
        DrawPlugin plugin = DrawPlugin.getCurrent();
        return (plugin == null) ? NullDrawingEditor.INSTANCE
                                : plugin.getDrawingEditor();
    }

    public void execute() {
        DrawingView view = getEditor().view();
        JDialog dialog = getDialog();
        Object value = null;
        FigureEnumeration k = view.selectionElements();
        while (k.hasMoreElements()) {
            Figure f = k.nextFigure();
            Object val = f.getAttribute(fAttribute);
            if (val != null) {
                if (value != null && !value.equals(val)) {
                    // different values, use previous value
                    value = null;
                    break;
                }
                value = val;
            }
        }
        String valuestr = "";
        if (value != null) {
            if (value instanceof Color) {
                valuestr = "0x"
                           + Integer.toHexString(((Color) value).getRGB())
                                    .substring(2);
            } else {
                valuestr = value.toString();
            }
        }
        input.setText(valuestr);
        resetFocus();
        dialog.show();
    }

    private void resetFocus() {
        input.requestFocus();
        input.setSelectionStart(0);
        input.setSelectionEnd(input.getText().length());
    }

    public boolean isExecutable() {
        if (getEditor() == NullDrawingEditor.INSTANCE) {
            return false;
        }
        return getEditor().view().selectionCount() > 0;
    }

    private boolean apply(Class type) {
        String valuestr = input.getText();
        Object value = null;
        try {
            if (type == Integer.class) {
                value = Integer.valueOf(valuestr);
            } else if (type == String.class) {
                value = valuestr;
            } else if (type == Color.class) {
                value = Color.decode(valuestr);
            }
        } catch (NumberFormatException ex) {
        }
        if (value != null) {
            new ChangeAttributeCommand("", fAttribute, value).execute();
        }
        resetFocus();
        return value != null;
    }
}