package de.renew.formalism.base;

import de.renew.net.Net;
import de.renew.net.NetLoader;
import de.renew.net.Place;
import de.renew.net.Transition;
import de.renew.net.TransitionInscription;
import de.renew.net.UplinkInscription;

import de.renew.shadow.ShadowArc;
import de.renew.shadow.ShadowCompiler;
import de.renew.shadow.ShadowDeclarationNode;
import de.renew.shadow.ShadowInscription;
import de.renew.shadow.ShadowLookup;
import de.renew.shadow.ShadowNet;
import de.renew.shadow.ShadowNetElement;
import de.renew.shadow.ShadowPlace;
import de.renew.shadow.ShadowPreprocessor;
import de.renew.shadow.ShadowTransition;
import de.renew.shadow.SyntaxException;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;


public abstract class AbstractSingleNetCompiler implements ShadowCompiler {
    protected ShadowLookup lookup;
    protected NetLoader loopbackNetLoader;
    protected transient int placeNum = 0;
    protected transient int transitionNum = 0;

    public void setShadowLookup(ShadowLookup lookup) {
        this.lookup = lookup;
    }

    public ShadowLookup getLookup() {
        return lookup;
    }

    public void setLoopbackNetLoader(NetLoader loopbackNetLoader) {
        this.loopbackNetLoader = loopbackNetLoader;
    }

    /**
     * Create a single net with the given name.
     * Override to customize the created net.
     **/
    public Net createNet(String name) {
        return new Net(name);
    }

    public ShadowPreprocessor[] getRequiredPreprocessors() {
        return new ShadowPreprocessor[0];
    }

    protected abstract void compilePlaceInscriptions(ShadowPlace shadowPlace,
                                                     Place place)
            throws SyntaxException;

    // Do the actual compilation.
    protected void compile(ShadowPlace shadowPlace, Net net)
            throws SyntaxException {
        // Determine the name.
        String pname = shadowPlace.getName();
        if (pname == null) {
            placeNum++;
            pname = "P" + placeNum;
        }

        // Create the new place;
        Place place = new Place(net, pname);
        place.getID().setFigureID(shadowPlace.getID());
        place.setTrace(shadowPlace.getTrace());
        lookup.set(shadowPlace, place);

        compilePlaceInscriptions(shadowPlace, place);
    }

    protected abstract void compileArc(ShadowArc shadowArc)
            throws SyntaxException;

    protected abstract Collection /*TransitionInscription*/ compileTransitionInscription(ShadowInscription inscription)
            throws SyntaxException;

    protected void compileNonStandardTransitionInscription(ShadowNetElement elem)
            throws SyntaxException {
        throw new SyntaxException("Unsupported type of transition inscription",
                                  null).addObject(elem);
    }

    protected void compileTransitionInscriptions(ShadowTransition shadowTransition,
                                                 Vector parsedInscriptions,
                                                 Vector errorShadows)
            throws SyntaxException {
        // The transition already exists.
        // Only insert the inscriptions.
        Iterator inscriptions = shadowTransition.elements().iterator();
        while (inscriptions.hasNext()) {
            Object elem = inscriptions.next();
            if (elem instanceof ShadowInscription) {
                ShadowInscription inscription = (ShadowInscription) elem;

                Iterator subinscriptions = null;
                try {
                    subinscriptions = compileTransitionInscription((ShadowInscription) elem)
                                          .iterator();
                } catch (SyntaxException e) {
                    throw e.addObject(inscription);
                }

                while (subinscriptions.hasNext()) {
                    parsedInscriptions.addElement(subinscriptions.next());
                    errorShadows.addElement(inscription);
                }
            } else if (elem instanceof ShadowArc) {
                compileArc((ShadowArc) elem);
            } else if (elem instanceof ShadowNetElement) {
                compileNonStandardTransitionInscription((ShadowNetElement) elem);
            } else {
                throw new SyntaxException("Unsupported type of transition inscription",
                                          null).addObject(elem);
            }
        }
    }

    protected void compile(ShadowTransition shadowTransition, Net net)
            throws SyntaxException {
        // Determine the name.
        String tname = shadowTransition.getName();
        if (tname == null) {
            transitionNum++;
            tname = "T" + transitionNum;
        }

        // Create the new transition.
        Transition transition = new Transition(net, tname);
        transition.getID().setFigureID(shadowTransition.getID());
        transition.setTrace(shadowTransition.getTrace());
        lookup.set(shadowTransition, transition);

        // Prepare a list to hold the inscriptions.
        Vector parsedInscriptions = new Vector();
        Vector errorShadows = new Vector();


        // Insert the inscriptions.
        compileTransitionInscriptions(shadowTransition, parsedInscriptions,
                                      errorShadows);

        // Check for multiple uplinks.
        int uplinkCount = 0;
        for (int i = 0; i < parsedInscriptions.size(); i++) {
            if (parsedInscriptions.elementAt(i) instanceof UplinkInscription) {
                uplinkCount++;
            }
        }

        if (uplinkCount > 1) {
            SyntaxException e = new SyntaxException("Transition has more than one uplink.",
                                                    null);
            for (int i = 0; i < parsedInscriptions.size(); i++) {
                if (parsedInscriptions.elementAt(i) instanceof UplinkInscription) {
                    e.addObject(errorShadows.elementAt(i));
                }
            }
            throw e;
        }

        // Insert inscriptions.
        for (int i = 0; i < parsedInscriptions.size(); i++) {
            transition.add((TransitionInscription) parsedInscriptions.elementAt(i));
        }
    }

    public abstract void parseDeclarations(ShadowNet shadowNet)
            throws SyntaxException;

    /**
     * Return the single declaration node of the shadow net.
     * Throw an error if more than one declaration node exists.
     * Return null, if no declaration node could be found.
     */
    protected ShadowDeclarationNode findDeclarationNode(ShadowNet shadowNet)
            throws SyntaxException {
        // Search for declarations.
        Iterator iterator = shadowNet.elements().iterator();
        ShadowDeclarationNode node = null;
        while (iterator.hasNext()) {
            Object elem = iterator.next();
            if (elem instanceof ShadowDeclarationNode) {
                // Have we already found a declaration node?
                if (node != null) {
                    SyntaxException e = new SyntaxException("Only one declaration node is allowed.",
                                                            null);
                    e.addObject(node);
                    e.addObject((ShadowDeclarationNode) elem);
                    while (iterator.hasNext()) {
                        elem = iterator.next();
                        if (elem instanceof ShadowDeclarationNode) {
                            e.addObject((ShadowDeclarationNode) elem);
                        }
                    }
                    throw e;
                }


                // Remember the declaration node we found.
                node = (ShadowDeclarationNode) elem;
            }
        }

        return node;
    }

    public void compile(ShadowNet shadowNet) throws SyntaxException {
        // Find the net that is compiled.
        Net net = lookup.getNet(shadowNet.getName());

        // Compile the individual net elements.
        // First the declaration node, if any.
        parseDeclarations(shadowNet);
        // Second the places.
        Iterator iterator = shadowNet.elements().iterator();
        while (iterator.hasNext()) {
            Object elem = iterator.next();
            if (elem instanceof ShadowPlace) {
                compile((ShadowPlace) elem, net);
            }
        }


        // Third the transitions.
        iterator = shadowNet.elements().iterator();
        while (iterator.hasNext()) {
            Object elem = iterator.next();
            if (elem instanceof ShadowTransition) {
                compile((ShadowTransition) elem, net);
            }
        }
    }
}