package de.renew.formalism.java;

import de.renew.application.SequentialOnlyExtension;

import de.renew.expression.ConstantExpression;
import de.renew.expression.Expression;
import de.renew.expression.VariableExpression;

import de.renew.formalism.base.AbstractSingleNetCompiler;

import de.renew.net.ExpressionTokenSource;
import de.renew.net.Net;
import de.renew.net.Place;
import de.renew.net.Transition;
import de.renew.net.UplinkInscription;
import de.renew.net.arc.Arc;
import de.renew.net.arc.ClearArc;
import de.renew.net.inscription.ActionInscription;
import de.renew.net.inscription.DownlinkInscription;
import de.renew.net.inscription.ExpressionInscription;
import de.renew.net.inscription.GuardInscription;

import de.renew.shadow.ShadowArc;
import de.renew.shadow.ShadowDeclarationNode;
import de.renew.shadow.ShadowInscription;
import de.renew.shadow.ShadowLookup;
import de.renew.shadow.ShadowNet;
import de.renew.shadow.ShadowPlace;
import de.renew.shadow.SyntaxException;

import de.renew.unify.Tuple;

import de.renew.util.Types;

import java.io.StringReader;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;


public class SingleJavaNetCompiler extends AbstractSingleNetCompiler {
    private boolean allowDangerousArcs;
    private boolean allowTimeInscriptions;
    private boolean wantEarlyTokens;
    private transient Hashtable placeTypes = new Hashtable(); // Maps shadow places to types.
    protected ParsedDeclarationNode declaration;

    public SingleJavaNetCompiler(boolean allowDangerousArcs,
                                 boolean allowTimeInscriptions,
                                 boolean wantEarlyTokens) {
        this.allowDangerousArcs = allowDangerousArcs;
        this.allowTimeInscriptions = allowTimeInscriptions;
        this.wantEarlyTokens = wantEarlyTokens;
    }

    public Net createNet(String name) {
        Net net = super.createNet(name);
        net.setEarlyTokens(wantEarlyTokens);
        return net;
    }

    protected InscriptionParser makeParser(String inscr) {
        JavaNetParser parser = new JavaNetParser(new StringReader(inscr));
        parser.setNetLoader(loopbackNetLoader);
        return parser;
    }

    public static SyntaxException makeSyntaxException(ParseException e) {
        Token t = e.currentToken;
        if (t.next != null) {
            t = t.next;
        }
        return new SyntaxException(e.getMessage(), null, t.beginLine,
                                   t.beginColumn);
    }

    // Check the validity of certain inscriptions.
    public String checkDeclarationNode(String inscr, boolean special,
                                       ShadowNet shadowNet)
            throws SyntaxException {
        parseDeclarationNode(inscr);
        return "declaration";
    }

    public String checkArcInscription(String inscr, boolean special,
                                      ShadowNet shadowNet)
            throws SyntaxException {
        makeDeclarationNode(shadowNet);
        parseArcInscription(inscr);
        return "inscription";
    }

    public String checkTransitionInscription(String inscr, boolean special,
                                             ShadowNet shadowNet)
            throws SyntaxException {
        makeDeclarationNode(shadowNet);
        Collection inscriptions = makeInscriptions(inscr, null, false);

        if (inscriptions.size() != 1) {
            return null;
        }

        Object res = inscriptions.iterator().next();
        if (res instanceof ActionInscription) {
            return "action";
        } else if (res instanceof GuardInscription) {
            return "guard";
        } else if (res instanceof UplinkInscription) {
            return "uplink";
        } else if (res instanceof DownlinkInscription) {
            return "downlink";
        } else if (res instanceof ExpressionInscription) {
            return "expression";
        } else {
            return null;
        }
    }

    public String checkPlaceInscription(String inscr, boolean special,
                                        ShadowNet shadowNet)
            throws SyntaxException {
        makeDeclarationNode(shadowNet);
        Collection placeinscriptions = parsePlaceInscription(inscr);

        if (placeinscriptions.size() != 1) {
            return null;
        }

        Object element = placeinscriptions.iterator().next();
        if (element instanceof Class) {
            return "type";
        } else if (element instanceof PlaceBehaviourModifier) {
            return "placebehaviour";
        } else {
            return "initialMarking";
        }
    }

    protected void compilePlaceInscriptions(ShadowPlace shadowPlace, Place place)
            throws SyntaxException {
        // Store the initialisation inscriptions for later use.
        Vector parsedInscriptions = new Vector();
        Vector errorShadows = new Vector();

        // Insert the inscriptions.
        Iterator iterator = shadowPlace.elements().iterator();
        while (iterator.hasNext()) {
            Object elem = iterator.next();
            if (elem instanceof ShadowInscription) {
                String inscr = ((ShadowInscription) elem).inscr;

                try {
                    Iterator exprEnum = parsePlaceInscription(inscr).iterator();
                    while (exprEnum.hasNext()) {
                        parsedInscriptions.addElement(exprEnum.next());
                        errorShadows.addElement(elem);
                    }
                } catch (SyntaxException e) {
                    throw e.addObject((ShadowInscription) elem);
                }
            } else if (elem instanceof ShadowArc) {
            } else {
                throw new SyntaxException("Unsupported place inscription", null).addObject(shadowPlace)
                                                                                .addObject(elem);
            }
        }

        // Remember the type of this place.
        Class type = Types.UNTYPED;
        int typeCount = 0;
        PlaceBehaviourModifier behaviour = new PlaceBehaviourModifier(Place.MULTISETPLACE);
        int behaviourCount = 0;

        for (int i = 0; i < parsedInscriptions.size(); i++) {
            if (parsedInscriptions.elementAt(i) instanceof Class) {
                type = (Class) parsedInscriptions.elementAt(i);
                typeCount++;
            } else if (parsedInscriptions.elementAt(i) instanceof PlaceBehaviourModifier) {
                behaviour = (PlaceBehaviourModifier) parsedInscriptions
                            .elementAt(i);
                behaviourCount++;
            }
        }

        // Check for multiple types.
        if (typeCount > 1) {
            SyntaxException e = new SyntaxException("Place is typed more than once.",
                                                    null);
            for (int i = 0; i < parsedInscriptions.size(); i++) {
                if (parsedInscriptions.elementAt(i) instanceof Class) {
                    e.addObject(errorShadows.elementAt(i));
                }
            }
            throw e;
        }

        // Check for multiple behaviours.
        if (behaviourCount > 1) {
            SyntaxException e = new SyntaxException("Place has more than one assigned behaviour.",
                                                    null);
            for (int i = 0; i < parsedInscriptions.size(); i++) {
                if (parsedInscriptions.elementAt(i) instanceof PlaceBehaviourModifier) {
                    e.addObject(errorShadows.elementAt(i));
                }
            }
            throw e;
        }

        // Remember the type information for this place.
        if (type != Types.UNTYPED) {
            placeTypes.put(shadowPlace, type);
        }


        // Set the behaviour information for this place.
        place.setBehaviour(behaviour.behaviour);

        // Now add the initial markings.
        for (int i = 0; i < parsedInscriptions.size(); i++) {
            if (parsedInscriptions.elementAt(i) instanceof TypedExpression) {
                TypedExpression expr = (TypedExpression) parsedInscriptions
                                       .elementAt(i);

                Expression castedExpression = null;
                try {
                    castedExpression = JavaNetHelper.makeCastedOutputExpression(type,
                                                                                expr);
                } catch (SyntaxException e) {
                    throw e.addObject(errorShadows.elementAt(i));
                }
                place.add(new ExpressionTokenSource(castedExpression));
            }
        }
    }

    Class getType(ShadowPlace place) {
        if (placeTypes.containsKey(place)) {
            return (Class) placeTypes.get(place);
        } else {
            return Types.UNTYPED;
        }
    }

    protected void compileArc(ShadowArc shadowArc) throws SyntaxException {
        ArcFactory factory = getArcFactory(shadowArc);
        if (shadowArc.shadowArcType == ShadowArc.doubleHollow) {
            compileClearArc(shadowArc);
        } else {
            compileArcInscriptions(shadowArc, factory);
        }
    }

    protected Collection compileTransitionInscription(ShadowInscription inscription)
            throws SyntaxException {
        return makeInscriptions(inscription.inscr, lookup, true);
    }

    protected ShadowInscription getSingleArcInscription(ShadowArc arc)
            throws SyntaxException {
        Set inscriptions = arc.elements();
        if (inscriptions.size() >= 2) {
            SyntaxException e = new SyntaxException("This arc must not be inscribed "
                                                    + "multiple times.", null)
                                .addObject(arc);
            Iterator iterator = inscriptions.iterator();
            while (iterator.hasNext()) {
                e.addObject((ShadowInscription) iterator.next());
            }
            throw e;
        } else if (inscriptions.size() == 1) {
            return (ShadowInscription) inscriptions.iterator().next();
        } else {
            return null;
        }
    }

    protected TypedExpression parseVariable(String inscr)
            throws SyntaxException {
        InscriptionParser parser = makeParser(inscr);
        parser.setDeclarationNode(declaration);
        try {
            return parser.VariableInscription();
        } catch (ParseException e) {
            throw makeSyntaxException(e);
        }
    }

    protected void compileClearArc(ShadowArc shadowArc)
            throws SyntaxException {
        Transition transition = lookup.get(shadowArc.transition);
        Place place = lookup.get(shadowArc.place);
        Class placeType = getType(shadowArc.place);

        ShadowInscription shadowInscription = getSingleArcInscription(shadowArc);
        String inscr = null;
        if (shadowInscription != null) {
            inscr = shadowInscription.inscr;
        }

        if (inscr == null || inscr.equals("")) {
            // For empty clear arcs, prepare an anonymous variable.
            ClearArc arc = new ClearArc(place,
                                        new VariableExpression(Types.UNTYPED,
                                                               null),
                                        Object.class);
            arc.setTrace(shadowArc.getTrace());
            transition.add(arc);
        } else {
            // We have got a expression to parse.
            TypedExpression expr;
            try {
                expr = parseVariable(inscr);
            } catch (SyntaxException e) {
                throw e.addObject(shadowInscription);
            }
            Class exprType = expr.getType();

            Class elementType;
            if (placeType == Types.UNTYPED) {
                if (exprType != Types.UNTYPED) {
                    throw new SyntaxException("Cannot clear untyped place using "
                                              + "typed variable.", null)
                          .addObject(shadowArc);
                }


                // If primitive values happen to occur in the
                // place, they are wrapped in Value objects.
                //
                // This exposes the Value class to users, so maybe we should
                // complain? Or should we disable the transition at runtime,
                // if primitive values occur in the place?
                elementType = Object.class;
            } else {
                if (exprType == Types.UNTYPED) {
                    elementType = placeType;
                } else if (!exprType.isArray()) {
                    throw new SyntaxException("Variable of array type expected.",
                                              null).addObject(shadowInscription);
                } else {
                    elementType = exprType.getComponentType();

                    if (!Types.allowsLosslessWidening(placeType, elementType)) {
                        throw new SyntaxException("Cannot losslessly convert "
                                                  + JavaHelper
                                                    .makeTypeErrorString(placeType)
                                                  + " to "
                                                  + JavaHelper
                                                    .makeTypeErrorString(elementType)
                                                  + ".", null).addObject(shadowInscription);
                    }
                }
            }

            ClearArc arc = new ClearArc(place, expr.getExpression(), elementType);
            arc.setTrace(shadowArc.getTrace());
            transition.add(arc);
        }
    }

    // Returns an enumeration of correctly timed expressions.
    // This method does not do a lot. Simply checking for
    // void expressions now. Possible simplification.
    protected Collection makeArcExpressions(String inscription)
            throws SyntaxException {
        Iterator timedExprs = parseArcInscription(inscription).iterator();
        List seq = new ArrayList();

        while (timedExprs.hasNext()) {
            TimedExpression expr = (TimedExpression) timedExprs.next();

            if (expr.getExpression().getType() == Void.TYPE) {
                throw new SyntaxException("Cannot use void expressions as arc inscriptions.",
                                          null);
            }
            if (expr.isTimed()) {
                setSequential(true);
            }

            seq.add(expr);
        }

        return seq;
    }

    // Returns an enumeration of typed expressions.
    protected Collection makeEmptyArcExpressions() throws SyntaxException {
        return makeArcExpressions(null);
    }

    protected void compileSingleArcInscription(Collection exprs,
                                               Class placeType,
                                               ArcFactory factory, Place place,
                                               Transition transition,
                                               boolean trace)
            throws SyntaxException {
        Iterator iterator = exprs.iterator();
        while (iterator.hasNext()) {
            TimedExpression timedExpr = (TimedExpression) iterator.next();

            if (factory.allowsTime() && allowTimeInscriptions) {
                // Verify that the time expression is numeric.
                if (timedExpr.isTimed() && timedExpr.getTime().isTyped()) {
                    if (!Types.allowsWideningConversion(timedExpr.getTime()
                                                                         .getType(),
                                                                Double.TYPE)) {
                        throw new SyntaxException("Non-numeric time expression.",
                                                  null);
                    }
                }
            } else if (timedExpr.isTimed()) {
                if (allowTimeInscriptions) {
                    throw new SyntaxException("This arc type does not allow time inscriptions.",
                                              null);
                } else {
                    throw new SyntaxException("Time annotations are not allowed.",
                                              null);
                }
            }

            factory.compileArc(place, transition, trace, placeType, timedExpr);
        }
    }

    protected void compileArcInscriptions(ShadowArc shadowArc,
                                          ArcFactory factory)
            throws SyntaxException {
        Transition transition = lookup.get(shadowArc.transition);
        Place place = lookup.get(shadowArc.place);

        Iterator inscriptions = shadowArc.elements().iterator();
        if (inscriptions.hasNext()) {
            do {
                ShadowInscription inscription = (ShadowInscription) inscriptions
                                                .next();

                // Create an individual arc for each inscription.
                try {
                    compileSingleArcInscription(makeArcExpressions(inscription.inscr),
                                                getType(shadowArc.place),
                                                factory, place, transition,
                                                shadowArc.getTrace());
                } catch (SyntaxException e) {
                    throw e.addObject(inscription);
                }
            } while (inscriptions.hasNext());
        } else {
            // Create a single arc with the default inscription.
            try {
                factory.emptyArcCheck();

                compileSingleArcInscription(makeEmptyArcExpressions(),
                                            getType(shadowArc.place), factory,
                                            place, transition,
                                            shadowArc.getTrace());
            } catch (SyntaxException e) {
                throw e.addObject(shadowArc);
            }
        }
    }

    protected ArcFactory getArcFactory(int arcType, boolean allowTime) {
        return new SimpleArcFactory(arcType, allowTime);
    }

    // Redefine this method to map some or all shadow arc types to
    // different simulator arc types.
    protected ArcFactory getArcFactory(ShadowArc shadowArc)
            throws SyntaxException {
        switch (shadowArc.shadowArcType) {
        case ShadowArc.test:
            return getArcFactory(Arc.test, false);
        case ShadowArc.both:
            return getArcFactory(Arc.bothOT, true);
        case ShadowArc.ordinary:
            return getArcFactory(shadowArc.placeToTransition ? Arc.in : Arc.out,
                                 true);
        case ShadowArc.inhibitor:
            //            if (allowDangerousArcs) {
            setSequential(true);
            return getArcFactory(Arc.inhibitor, false);

        //            } else {
        //            	throw new SyntaxException("InhibitorArcs not supported in parallel Simulation setting."); 
        //            }
        //            break;
        case ShadowArc.doubleOrdinary:
            if (shadowArc.placeToTransition) {
                return FlexibleInArcFactory.INSTANCE;
            } else {
                return FlexibleOutArcFactory.INSTANCE;
            }
        case ShadowArc.doubleHollow:
            //            if (allowDangerousArcs) {
            setSequential(true);
            return null; // special case: there is no factory!
        }
        throw new SyntaxException("Unsupported arc type.", null).addObject(shadowArc);
    }

    Collection parsePlaceInscription(String inscr) throws SyntaxException {
        if ((inscr != null) && !inscr.equals("")) {
            InscriptionParser parser = makeParser(inscr);
            parser.setDeclarationNode(declaration);
            try {
                return parser.PlaceInscription();
            } catch (ParseException e) {
                // I really do not know what this could be.
                // I'll give out the error message of the
                // second parse, although it might be
                // appropriate to tell the user that I also
                // tried to parse a type.
                throw makeSyntaxException(e);
            }
        }

        return Collections.EMPTY_SET;
    }

    // Returns an enumeration of typed expressions.
    protected Collection parseArcInscription(String inscr)
            throws SyntaxException {
        if (inscr != null && !inscr.equals("")) {
            InscriptionParser parser = makeParser(inscr);
            parser.setDeclarationNode(declaration);
            try {
                return parser.ArcInscription();
            } catch (ParseException e) {
                throw makeSyntaxException(e);
            }
        } else {
            return Collections.singleton(new TimedExpression(new TypedExpression(Tuple.class,
                                                                                 new ConstantExpression(Tuple.class,
                                                                                                        new Tuple(0))),
                                                             null));
        }
    }

    protected Collection makeInscriptions(String str, ShadowLookup lookup,
                                          boolean create)
            throws SyntaxException {
        // The input string must be parsed, it denotes either a
        // creation inscription, an uplink, a downlink, ...
        //
        // Syntax:
        //   x:new NetName          creation
        //   x:channel(...)         downlink
        //   :channel(...)          uplink
        //   x=new a.java.Type(...) creation
        //   x=y.method(...)        call
        //   guard ...              guard
        //   action x= ...          action
        InscriptionParser parser = makeParser(str);
        parser.setLookup(lookup);
        parser.setDeclarationNode(declaration);

        try {
            return parser.TransitionInscription(create);
        } catch (ParseException e) {
            throw makeSyntaxException(e);
        }
    }

    // declaration node processing
    protected ParsedDeclarationNode compile(ShadowDeclarationNode declaration)
            throws SyntaxException {
        try {
            return parseDeclarationNode(declaration.inscr);
        } catch (SyntaxException e) {
            throw e.addObject(declaration);
        }
    }

    ParsedDeclarationNode parseDeclarationNode(String inscr)
            throws SyntaxException {
        if (inscr != null) {
            InscriptionParser parser = makeParser(inscr);
            try {
                return parser.DeclarationNode();
            } catch (ParseException e) {
                throw makeSyntaxException(e);
            }
        } else {
            return makeEmptyDeclarationNode(null);
        }
    }

    protected ParsedDeclarationNode makeEmptyDeclarationNode(ShadowNet net) {
        return new ParsedDeclarationNode();
    }

    public void parseDeclarations(ShadowNet shadowNet)
            throws SyntaxException {
        makeDeclarationNode(shadowNet);
    }

    public ParsedDeclarationNode makeDeclarationNode(ShadowNet shadowNet)
            throws SyntaxException {
        // Compile the declaration node.
        ShadowDeclarationNode node = findDeclarationNode(shadowNet);
        if (node == null) {
            declaration = makeEmptyDeclarationNode(shadowNet);
        } else {
            declaration = compile((ShadowDeclarationNode) node);
        }
        return declaration;
    }

    private void setSequential(boolean value) {
        SequentialOnlyExtension seqEx = SequentialOnlyExtension.lookup(lookup);
        seqEx.setSequentialOnly(value);
    }
}