package de.renew.gui;

import de.renew.application.IllegalCompilerException;
import de.renew.application.SimulationEnvironment;
import de.renew.application.SimulatorPlugin;

import de.renew.engine.searcher.ConsoleTracerFactory;
import de.renew.engine.searcher.TracerFactory;
import de.renew.engine.simulator.Simulator;

import de.renew.net.NetInstance;
import de.renew.net.NetNotFoundException;

import de.renew.plugin.PluginProperties;

import de.renew.remote.NetInstanceAccessor;

import de.renew.shadow.ShadowCompilerFactory;
import de.renew.shadow.ShadowLookup;
import de.renew.shadow.ShadowNetSystem;
import de.renew.shadow.SyntaxException;

import java.util.ConcurrentModificationException;
import java.util.Iterator;


public class CPNSimulation {
    private ShadowNetSystem netSystem;
    private Simulator simulator = null;
    private boolean sequentialOnly;
    private BreakpointManager breakpointManager = null;
    private CPNDrawingLoader drawingLoader;
    private SimulatorPlugin simulatorPlugin;

    public CPNSimulation(boolean onlySequential, CPNDrawingLoader loader) {
        this.simulatorPlugin = SimulatorPlugin.getCurrent();
        this.sequentialOnly = onlySequential;
        this.drawingLoader = loader;
        newNetSystem();
    }

    public boolean isStrictlySequential() {
        return sequentialOnly;
    }

    public void buildAllShadows() {
        newNetSystem();
        boolean again = true;
        while (again) {
            again = false;
            Iterator drawings = drawingLoader.loadedDrawings();

            try {
                // Build new shadows of all nets and all elements.
                // Do everything in a fine order:
                while (drawings.hasNext()) {
                    CPNDrawing currentDrawing = (CPNDrawing) drawings.next();
                    if (PluginProperties.debug()) {
                        System.out.println("CPNSimulation: Building shadow for drawing "
                                           + currentDrawing + ".");
                    }
                    currentDrawing.buildShadow(netSystem);
                }
            } catch (ConcurrentModificationException e) {
                System.err.println("CPNSimulation.buildAllShadows(): Concurrent modification. Redo from start...");
                again = true;
            }
        }
    }

    /**
     * Creates a new ShadowNetSystem, sets the information about
     * compiler and net loader, and updates the private instance
     * variable {@link #netSystem}.
     **/
    private void newNetSystem() {
        try {
            simulatorPlugin.possiblySetupClassSource(simulatorPlugin
                                                     .getProperties());
        } catch (IllegalStateException e) {
            System.err.println("CPNSimulation: Cannot configure classReinit mode while a simulation is running.");
            if (PluginProperties.debug()) {
                e.printStackTrace();
            }
        }

        ShadowCompilerFactory compilerFactory = ModeReplacement.getInstance()
                                                               .getDefaultCompilerFactory();
        if (compilerFactory == null) {
            System.err.println("CPNSimulation: cannot start, got no compiler.");
        }

        netSystem = new ShadowNetSystem(compilerFactory,
                                        new GuiShadowNetLoader(drawingLoader,
                                                               simulatorPlugin
                                                               .getProperties()));
        if (PluginProperties.debug()) {
            System.out.println("CPNSimulation: New net system created.");
        }
    }

    public ShadowNetSystem getNetSystem() {
        // Make sure that all shadows are correctly set up.
        buildAllShadows();
        // Return the shadow net system.
        return netSystem;
    }

    CPNDrawingLoader getDrawingLoader() {
        return drawingLoader;
    }

    void setBreakpointManager(BreakpointManager newBreakpointManager) {
        BreakpointManager oldBreakpointManager = breakpointManager;
        breakpointManager = newBreakpointManager;
        if (oldBreakpointManager != null) {
            simulatorPlugin.removeExtension(oldBreakpointManager);
        }
        simulatorPlugin.addExtension(newBreakpointManager);
    }

    /**
     * Returns the actual breakpoint manager.
     * Attention: May be <code>null</code>!
     */
    public BreakpointManager getBreakpointManager() {
        return breakpointManager;
    }

    public SimulatorPlugin getSimulatorPlugin() {
        return simulatorPlugin;
    }

    public void syntaxCheckOnly() throws SyntaxException {
        // Make sure to recreate all shadows, so that changes of
        // the drawings are incorporated.
        buildAllShadows();

        // Compile and flatten the lookup table.
        ShadowLookup lookup = netSystem.compile();
    }

    /**
     * Start a new simulation.
     * Return the initially created NetInstance,
     * if the creation of the initial instance
     * was successful, or null, if no initial creation took
     * place.
     *
     * @param mainNet  the name of the net to create the first
     *                 net instance from.
     *
     * @throws SyntaxException
     *   if the compiler found any errors in any net drawing.
     *
     * @throws NetNotFoundException
     *   if no net with the given name <code>mainNet</code> exists.
     */
    public NetInstanceAccessor initSimulation(String mainNet)
            throws SyntaxException, NetNotFoundException {
        buildAllShadows();

        // Set the simulation up
        TracerFactory tracerFactory = new ConsoleTracerFactory("");
        simulatorPlugin.setDefaultNetLoader();
        simulatorPlugin.setupSimulation(null, tracerFactory);


        // Compile (the simulator plugin does that for us).
        // If something goes wrong, terminate the simulation.
        try {
            simulatorPlugin.insertNets(netSystem);

            NetInstanceAccessor primaryInstance = simulatorPlugin.wrapInstance(simulatorPlugin
                                                                               .createNetInstance(mainNet));
            if (primaryInstance == null) {
                simulatorPlugin.terminateSimulation();
            }
            return primaryInstance;
        } catch (SyntaxException e) {
            simulatorPlugin.terminateSimulation();
            throw e;
        } catch (NetNotFoundException e) {
            simulatorPlugin.terminateSimulation();
            throw e;
        } catch (IllegalCompilerException e) {
            simulatorPlugin.terminateSimulation();
            throw e;
        }
    }

    /**
     * Writes all currently known <code>NetInstance</code>s,
     * <code>Net</code>s, the <code>SearchQueue</code> contents
     * and some additional information to the stream. The written
     * information is sufficient to continue the simulation from
     * the same state after deserialization.
     * <p>
     * <b>Side effect:</b> Any open <code>BindingSelectionFrame</code>
     * will be closed on execution of this method!
     * </p>
     * <p>
     * With Renew 2.0, most of this method's functionality has moved
     * to the <code>SimulatorPlugin</code>.
     * </p>
     *
     * @param output target stream
     *
     * @see SimulatorPlugin#saveState
     **/
    public void saveState(java.io.ObjectOutput output)
            throws java.io.IOException {
        // We have to ensure that the simulation is stopped.
        // Yes, the simulator plugin enforces this, too, but we
        // need to freeze the state before we collect the local
        // net instances and close the selection frame.
        synchronized (simulatorPlugin) {
            simulatorPlugin.getCurrentEnvironment().getSimulator().stopRun();

            // Close the BindingSelectionFrame (otherwise it would
            // be stored with its transition).
            BindingSelectionFrame.close();

            // Collect all local instances from open drawings.
            NetInstance[] instances = CPNInstanceDrawing.getAllLocalInstances();

            // Let the simulator plugin save the state.
            simulatorPlugin.saveState(output, instances);
        }
    }

    /**
     * Restores a simulation saved by <code>saveState()</code>.
     * This method mostly delegates to {@link SimulatorPlugin#loadState}.
     * It additionally tries to reopen the explicitly named net
     * instances within <code>CPNInstanceDrawing</code>s.
     * <p>
     * With Renew 2.0, most of this method's functionality has moved
     * to the <code>SimulatorPlugin</code>.
     * </p>
     *
     * @param input source stream
     *
     * @see SimulatorPlugin#loadState
     */
    public void loadState(java.io.ObjectInput input)
            throws java.io.IOException, ClassNotFoundException {
        NetInstance[] instances = simulatorPlugin.loadState(input, null,
                                                            new ConsoleTracerFactory(""));


        // Now try to open CPNInstanceDrawings
        // for all net instances which were loaded.
        NetInstanceAccessor currentInstance;
        CPNApplication editor = GuiPlugin.getCurrent().getGui();
        for (int i = 0; i < instances.length; i++) {
            currentInstance = simulatorPlugin.wrapInstance(instances[i]);
            if (editor != null) {
                editor.openInstanceDrawing(currentInstance);
            }
        }
    }

    public boolean isSimulationActive() {
        return simulatorPlugin.isSimulationActive();
    }

    public void simulationTerminate() {
        simulatorPlugin.terminateSimulation();
    }

    /**
     * To be called after an external firing of a transition was initiated.
     **/
    public void simulationRefresh() {
        SimulationEnvironment env = simulatorPlugin.getCurrentEnvironment();
        Simulator simulator = (env == null) ? null : env.getSimulator();
        if (simulator != null) {
            simulator.refresh();
        }
    }

    public void simulationStop() {
        SimulationEnvironment env = simulatorPlugin.getCurrentEnvironment();
        Simulator simulator = (env == null) ? null : env.getSimulator();
        if (simulator != null) {
            simulator.stopRun();
        }
    }

    public void simulationRun() {
        SimulationEnvironment env = simulatorPlugin.getCurrentEnvironment();
        Simulator simulator = (env == null) ? null : env.getSimulator();

        // To avoid simulation concurrent to serialization
        synchronized (this) {
            if (breakpointManager != null) {
                breakpointManager.clearLog();
            }
            if (simulator != null) {
                simulator.startRun();
            }
        }
    }

    public int simulationStep() {
        SimulationEnvironment env = simulatorPlugin.getCurrentEnvironment();
        Simulator simulator = (env == null) ? null : env.getSimulator();

        // To avoid simulation concurrent to serialization
        synchronized (this) {
            if (breakpointManager != null) {
                breakpointManager.clearLog();
            }
            if (simulator != null) {
                return simulator.step();
            } else {
                return Simulator.statusDisabled;
            }
        }
    }
}