package de.renew.plugin;

import java.io.File;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;


/**
 * This ClassLoader is responsible for loading user defined classes.
 * To find classes it looks searches the repositories given in the property
 * <code>de.renew.classPath</code>..
 * To set additional repositories it allows adding classpath items
 * dynamically.
 */
public class BottomClassLoader extends URLClassLoader {

    /**
     * the name of the property used to set the ClassPath
     */
    public final static String CLASSPATH_PROP_NAME = "de.renew.classPath";


    /**
     * Creates an instance of the BottomClassLoader.
     *
     * @param urls the urls used to find class files
     * @param parent the parent classloader
     */
    public BottomClassLoader(URL[] urls, ClassLoader parent) {
        super(urls, parent);
    }

    /**
     * Creates an instance of the BottomClassLoader. To determine the urls
     * for class loading the value of the property <code>de.renew.classPath</code>
     * will be evaluated.
     *
     * @param parent the parent classloader
     */
    public BottomClassLoader(ClassLoader parent) {
        super(new URL[0], parent);


        // evaluate the value of environment variable de.renew.classPath
        // and add classpath urls
        String classpath = System.getProperty(CLASSPATH_PROP_NAME);

        if (classpath != null) {
            String[] parts = classpath.split(new String(new char[] { File.pathSeparatorChar }));
            for (int x = 0; x < parts.length; x++) {
                try {
                    // if the part of the classpath is a legal url we can parse
                    // and add it directly.
                    addURL(new URL(parts[x]));
                } catch (MalformedURLException e1) {
                    // The part of the classpath was no legal url. Possibly it is
                    // a path to a local file or directory. So we should try to
                    // interpret it this way. 
                    File path = new File(parts[x]);


                    // if a file or path exists, add the corresponding url to 
                    // the classpath
                    if (path.exists()) {
                        try {
                            addURL(path.toURL());
                        } catch (MalformedURLException e2) {
                        }
                    }
                }
            }
        }
    }

    /**
     * Add the given URL to the list where classes are sought for.
     */
    public void addURL(URL url) {
        if (PluginProperties.debug()) {
            System.out.println(this + " adding URL " + url);
        }
        super.addURL(url);
    }
}