/*
 * (c) Copyright 2001 MyCorporation.
 * All Rights Reserved.
 */
package de.renew.plugin;

import de.renew.plugin.load.SimplePluginLoader;

import java.io.IOException;
import java.io.InputStream;

import java.net.MalformedURLException;
import java.net.URL;


/**
 * This is a standard implementation of the IPlugin interface.
 * When instantiated with its home URL, it uses the SimplePluginLoader's
 * PluginConfigFinder to initialize its PluginProperties (i.e., it reads the plugin.cfg
 * entries).
 *
 * It provides constructors with a PluginProperties argument (for the SimplePluginLoader);
 * it can also be instantiated with a URL.
 * In this case, it calls the PluginConfigFinder of the SimplePluginLoader to
 * create a PluginProperties object for its configuration.
 */
public class PluginAdapter implements IPlugin {
    protected PluginProperties _properties;

    /**
     * Create an instance of this plugin.
     * The PluginException being thrown may contain
     * - a IOException
     * - a MalformedURLException
     */
    public PluginAdapter(URL location) throws PluginException {
        PluginProperties pluginProps = new PluginProperties(location);
        try {
            InputStream stream = SimplePluginLoader.PluginConfigFinder
                                 .getConfigInputStream(location);
            pluginProps.load(stream);
            stream.close();
        } catch (MalformedURLException e) {
            throw new PluginException("PluginAdapter constructor: trying to find properties",
                                      e);
        } catch (IOException e) {
            throw new PluginException("PluginAdapter constructor: trying to read properties",
                                      e);
        }

        _properties = pluginProps;
    }

    /**
     * Creates a PluginAdapter with the given PluginProperties.
     */
    public PluginAdapter(PluginProperties props) {
        _properties = props;
    }

    public String getName() {
        return _properties.getProperty("name");
    }

    public String toString() {
        return "plugin " + getName();
    }

    public void init() {
    }

    public boolean cleanup() {
        return true;
    }

    public boolean canShutDown() {
        if (PluginProperties.debug()) {
            System.out.println("asking if shutdown is possible.");
        }
        return true;
    }

    public String getProperty(String prop) {
        return _properties.getProperty(prop);
    }

    public PluginProperties getProperties() {
        return _properties;
    }

    protected void registerExitBlock() {
        if (PluginProperties.debug()) {
            System.out.println(getClass() + " blocking exit");
        }
        PluginManager.getInstance().blockExit(this);
    }

    protected void registerExitOk() {
        if (PluginProperties.debug()) {
            System.out.println(getClass() + " releasing exit block");
        }
        PluginManager.getInstance().exitOk(this);
    }
}