package de.renew.plugin;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

import java.net.URL;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.Properties;


/**
 * An extension of the java.util.Properties class used by the
 * Plugin Loader system.
 * It uses System variables to request property values, if present;
 * if no variable is set, the file ".renew.properties" is
 * used.
 * <p>
 * If there is no entry for the requested property in there either,
 * a default value should be provided by the Plugin
 * (usually configured in the plugin.cfg file).
 * It contains a number of properties specific to the plugin system,
 * "mainClass", "provides", "requires" and "name", that
 * can be requested by dedicated get method (i.e.,
 * getName() instead of getProperty("name")).
 * Additionally, typing of properties is allowed:
 * getBoolProperty and getIntProperty try to convert the String object
 * into the respective type.
 * <p>
 * PluginProperties that belong to a specific plugin (i.e., are not
 * the userProperties),
 * provide a list of property names that are used by this plugin.
 * To do this, all properties in the configuration information
 * are considered;
 * consequently, the plugin developer needs to put ALL properties
 * in there.
 * This additionally takes care that a proper set of default values
 * are available.
 */
public class PluginProperties extends Properties {
    private static PluginProperties _userProperties;
    private URL _sourceURL;

    /**
     * The keyword to use for known properties without default
     * value in the plugin's configuration file. These properties
     * will be known to the plugin, but not set initially.
     * <p>
     * The actual string is &quot;<code>&lt;not set&gt;</code>&quot;
     * (case and whitespace sensitive).
     * </p>
     * <p>
     * <b>Example:</b><br>
     * The configuration file contains:
     * <pre>
     *     de.renew.remote.serverClass = &lt;not set&gt;
     * </pre>
     * The methods of this object will return the following results
     * (unless some other source has overridden the default value):<br>
     * - <code>
     *   {@link #getProperty getProperty}(&quot;de.renew.remote.serverClass&quot;) = null
     *   </code><br>
     * - <code>
     *   {@link #containsKey containsKey}(&quot;de.renew.remote.serverClass&quot;) = false
     *   </code><br>
     * - <code>
     *   {@link #isKnownProperty isKnownProperty}(&quot;de.renew.remote.serverClass&quot;) = true
     *   </code><br>
     * </p>
     **/
    public static final String NULL_VALUE = "<not set>";
    Collection _knownProperties = new HashSet();

    /**
     * This constructor is used to create the PluginProperties with the
     * specified URL, with entries loaded from the given InputStream.
     * It will override these entries with settings from the userProperties
     * and the system properties.
     * @throws IOException if the properties cannot be loaded from the
     *         specified stream. If that happens, no settings will have been set.
     */
    public PluginProperties(URL source, InputStream loadFrom)
            throws IOException {
        _sourceURL = source;

        load(loadFrom);

        // Add all property names from plugin.cfg to the set of known
        // properties (except mandatory plugin properties).
        _knownProperties.addAll(keySet());
        _knownProperties.remove("name");
        _knownProperties.remove("mainClass");
        _knownProperties.remove("requires");
        _knownProperties.remove("provides");

        // Unset known properties marked as {@link #NULL_VALUE}
        removeNullProperties();

        // Add all properties from the configuration file and the command
        // line.
        putAll(getUserProperties());
        putAll(System.getProperties());
    }

    public PluginProperties(URL source) {
        // assume default properties as base for all queried properties
        super(getUserProperties());
        // override with currently set System properties
        putAll(System.getProperties());

        _sourceURL = source;
    }

    /**
     * this constructor is provided so the user properties can be created without
     * infinite recursion.
     *
     */
    protected PluginProperties() {
        super();
        putAll(System.getProperties());
    }

    /**
     * Removes all property entries with the special value {@link #NULL_VALUE}.
     * The result is like the affected properties would never have been set.
     * This method is intendend to be called during creation time, after
     * the known properties have been determined from the plugin configuration
     * file.
     **/
    private void removeNullProperties() {
        Iterator entries = entrySet().iterator();
        while (entries.hasNext()) {
            Entry entry = (Entry) entries.next();
            if (NULL_VALUE.equals(entry.getValue())) {
                entries.remove();
            }
        }
    }

    public URL getURL() {
        return _sourceURL;
    }

    public String getMainClass() {
        return getProperty("mainClass", "");
    }

    public String getName() {
        return getProperty("name", "unknown");
    }

    /**
     * Returns a list of the services provided by the plugin
     * to which this property object belongs.
     */
    public Collection getProvisions() {
        String result = getProperty("provides", "");
        return PropertyHelper.parseListString(result);
    }

    /**
     * Returns a list of the services required by the plugin
     * to which this property object belongs.
     */
    public Collection getRequirements() {
        String result = getProperty("requires", "");
        return PropertyHelper.parseListString(result);
    }

    /**
     * Returns an array containing the names of all known properties.
     * A property is <i>known</i>, if it was included in the initial
     * set of properties loaded from the <code>URL</code> given to
     * the constructor of this <code>PluginProperties</code> instance.
     * @return an array of property names.
     **/
    public String[] getKnownProperties() {
        return (String[]) _knownProperties.toArray(new String[_knownProperties
                                                              .size()]);
    }

    /**
     * Tells whether the given property name is known.
     * A property is <i>known</i>, if it was included in the initial
     * set of properties loaded from the <code>URL</code> given to
     * the constructor of this <code>PluginProperties</code> instance.
     * @param name  the property name for the query
     * @return <code>true</code>, if the property with the given
     *         <code>name</code> is known.
     **/
    public boolean isKnownProperty(String name) {
        return _knownProperties.contains(name);
    }

    public String toString() {
        return "properties of Plugin " + getName();
    }

    /**
     * Return a String showing name, URL, mainClass of the corresponding Plugin
     * along with its provisions and requirements.
     */
    public String toExtString() {
        String result = "properties of Plugin " + getName() + ":\n";
        result += "loaded from URL " + getURL() + "\n";
        result += "main class is " + getMainClass() + "\n";
        result += "provides " + CollectionLister.toString(getProvisions())
        + "\n";
        result += "requires " + CollectionLister.toString(getRequirements());
        return result;
    }

    /**
     * Return a PluginProperties object that represents user settings
     * that are not specific to one plugin.
     */
    public static PluginProperties getUserProperties() {
        if (_userProperties == null) {
            String homeDir = System.getProperty("user.home");
            File home = new File(homeDir + File.separatorChar
                                 + ".renew.properties");
            try {
                System.out.println("Reading properties from "
                                   + home.getCanonicalPath());
                _userProperties = new PluginProperties();
                _userProperties._sourceURL = home.toURL();
                _userProperties.load(new FileInputStream(home));
            } catch (FileNotFoundException e) {
                System.err.println("PluginProperties: user properties file "
                                   + home + " does not exist.");
            } catch (IOException e) {
                System.err.println("PluginProperties: error reading user properties file "
                                   + home + ": " + e);
            }

            // now override set values with system properties again
            _userProperties.putAll(System.getProperties());

            //    		System.out.println ("user home dir is " + homeDir);
        }
        return _userProperties;
    }

    public static boolean debug() {
        return !("false".equalsIgnoreCase(getUserProperties().getProperty("debug",
                                                                          "false")));
    }

    /**
     * Returns the property as a bool with a default of "false".
     *
     * If the property is set to an empty String, "true"
     * is returned. This is to reflect the fact that it is true that the property
     * is set.
     *
     * @param property the property to convert to boolean
     * @return the boolean value of the property, false if not set.
     * @see PropertyHelper#getBoolProperty(Properties, String)
     */
    public boolean getBoolProperty(String property) {
        return PropertyHelper.getBoolProperty(this, property);
    }

    /**
     * Returns the property as an int.
     * Returns the value of the "def" parameter if property is not set.
     * @param property the property to convert to int
     * @return the parsed int value of the property
     */
    public int getIntProperty(String property, int def) {
        return PropertyHelper.getIntProperty(this, property, def);
    }

    /**
     * Returns the property as an int.
     * Returns -1 if the property is not set or an error occurs
     * @return the int value of the given property
     */
    public int getIntProperty(String property) {
        return PropertyHelper.getIntProperty(this, property);
    }

    /**
     * Returns the property as a List.
     * @param propName the property to be retrieved.
     * @see PropertyHelper.getListProperty(Properties, String)
     */
    public List getListProperty(String propName) {
        return PropertyHelper.getListProperty(this, propName);
    }
}