package de.renew.watch;

import collections.CircularList;
import collections.CollectionEnumeration;
import collections.LinkedList;
import collections.UpdatableSeq;

import de.renew.engine.searcher.Searcher;
import de.renew.engine.searcher.Tracer;
import de.renew.engine.searcher.TriggerCollection;
import de.renew.engine.searcher.Triggerable;
import de.renew.engine.simulator.ExecuteFinder;

import de.renew.net.NetInstance;

import de.renew.plugin.PluginProperties;

import de.renew.unify.Impossible;
import de.renew.unify.Tuple;
import de.renew.unify.Variable;

import de.renew.util.Semaphor;


public class ChannelSupervisor implements Triggerable {

    /**
     * A supervisor that might not have informed its watcher
     * about the possible bindings correctly will be inserted
     * into this list. Its set of bindings will be cleaned
     * up by the search thread.
     *
     * Because supervisors may be inserted and remove from the dirty
     * list asynchronously, it is required to synchronize on the
     * list before modifying it.
     **/
    private static UpdatableSeq dirtySupervisors = new CircularList();
    private static UpdatableSeq allSupervisors = new LinkedList();

    /**
     * This semaphor counts the number of elements in the
     * dirty list. It controls the search process.
     **/
    private static Semaphor dirtySem = new Semaphor();

    static {
        new Thread("Channel supervisor thread") {
                public void run() {
                    if (PluginProperties.debug()) {
                        System.out.println("ChannelSupervisor: Thread started.");
                    }
                    while (true) {
                        // Wait until the dirty list is non-empty.
                        dirtySem.P();
                        // Extract one element.
                        ChannelSupervisor supervisor;
                        synchronized (allSupervisors) {
                            supervisor = (ChannelSupervisor) dirtySupervisors
                                         .first();
                            dirtySupervisors.removeFirst();
                        }


                        // Pass control to that supervisor for an update.
                        supervisor.update();
                    }
                }
            }.start();
    }

    private final NetInstance netInstance;
    private final ChannelWatcher watcher;
    private final String channel;
    private final Class lateClass;
    private TriggerCollection triggers;

    public ChannelSupervisor(NetInstance netInstance, ChannelWatcher watcher,
                             String channel, Class lateClass) {
        this.netInstance = netInstance;
        this.watcher = watcher;
        this.channel = channel;
        this.lateClass = lateClass;

        triggers = new TriggerCollection(this);
        synchronized (allSupervisors) {
            allSupervisors.insertFirst(this);
        }

        proposeSearch();
    }

    // This method is currently unused.
    public boolean execute(Object early, Object late, Tracer tracer) {
        ExecuteFinder finder = new ExecuteFinder();
        Tuple paramTuple = new Tuple(new Object[] { early, late }, null);
        Variable variable = new Variable(paramTuple, null);

        Searcher searcher = new Searcher();
        searcher.initiatedSearch(netInstance, channel, variable, false, finder,
                                 null);

        boolean success = finder.isCompleted();
        if (success) {
            finder.execute(tracer, true);
        }

        return success;
    }

    public TriggerCollection triggers() {
        return triggers;
    }

    public void proposeSearch() {
        synchronized (allSupervisors) {
            dirtySupervisors.insertLast(this);
            triggers.clear();
        }
        dirtySem.V();
    }

    void update() {
        // Try to invoke the channel
        //   :channel(early,late);
        // where the early argument must be known after the search
        // and the late argument must be freely assignable.
        Tuple arguments = new Tuple(2);
        Variable earlyVariable = new Variable(arguments.getComponent(0), null);

        // We create a new searcher for this search process.
        Searcher searcher = new Searcher();


        // The searcher has to be rigged to satisfy the requirements
        // on the channels: the first channel must be completely known
        // after the search and the second channel must be freely assignable.
        try {
            // The first argument must be computed early
            // during the search for a binding.
            searcher.calcChecker.addEarlyVariable(earlyVariable, null);


            // The last argument must not be computed early.
            searcher.calcChecker.addCalculated(lateClass,
                                               arguments.getComponent(1), null,
                                               null);
        } catch (Impossible e) {
            throw new RuntimeException("Error in unification algorithm.");
        }


        // We must not reuse the old finder, because its set of
        // permissible values cannot be reset.
        ParamFinder finder = new ParamFinder(earlyVariable);


        // We want to find a transition within the specified
        // net instance that can fire.
        searcher.initiatedSearch(netInstance, channel,
                                 new Variable(arguments, null), false, finder,
                                 this);
        synchronized (this) {
            if (watcher != null) {
                watcher.bindingsCalculated(finder.elements());
            }
        }
    }

    private synchronized void discard() {
        if (watcher != null) {
            watcher.bindingsCalculated(de.renew.util.EmptyEnumeration.INSTANCE);
        }
    }

    public static void reset() {
        synchronized (allSupervisors) {
            dirtySupervisors.clear();
            CollectionEnumeration enum = allSupervisors.elements();
            while (enum.hasMoreElements()) {
                ChannelSupervisor supervisor = (ChannelSupervisor) enum
                                               .nextElement();
                supervisor.discard();
            }
            allSupervisors.clear();
        }
    }
}