package de.renew.ant;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.taskdefs.Delete;
import org.apache.tools.ant.taskdefs.Execute;
import org.apache.tools.ant.taskdefs.LogStreamHandler;
import org.apache.tools.ant.taskdefs.PumpStreamHandler;
import org.apache.tools.ant.types.FileSet;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;


/**
   A class that runs LaTeX and BIBTeX.<p>

   <b>Implementation plan:</b><br>
   <ol>

     <li> Execution of LaTeX and assorted programs by instantiating of
     <code>Execute</code>. An example could be <code>Javac</code>.

     <li> Deletion of intermediate files such as aux, toc, etc, via
     pattern sets, this is shown in <code>MatchingTask</code>. This
     easily allows addition of other pattern in the ant-file depending
     on the document.

     <li> Parsing the output of LaTeX should give hints, whether we
     need an additional run. Use a counter to abort from endless loops
     which may result from using the package <code>varioref</code>.

     <li> Test via junit at least the construction of the list of
     deletable files. Perhaps go further and test the command line or
     even the results of latex commands.

   </ol>
   @author Klaus Alfert <alfert@shire.prima.de>
   @version $Revision: 1.1 $ $Date: 2004/05/17 13:26:24 $
*/
public class LaTeX extends Task {
    // Attributes
    private String latexfile = null;
    private String bibtexfile = null;
    private File latexdir = null;
    private boolean verbose = true;
    private boolean bibtex = false;
    private boolean pdftex = false;
    private boolean psgztex = false;
    private boolean clean = false;
    private Delete delete = null;
    private final String[] deletePatterns = { "*.aux", "*.log", "*.toc", "*.lof", "*.lot", "*.bbl", "*.blg", ".out" };

    /** Constructor. Simply calls <code>super()</code>. */
    public LaTeX() {
        super();
    }

    /** Initialiazes some default-values which need a running environment. */
    public void init() {
        this.latexdir = this.project.getBaseDir();
    }

    /** The LaTeX main file*/
    public void setLatexfile(String filename) {
        latexfile = filename;
        bibtexfile = latexfile.substring(0, latexfile.lastIndexOf(".tex"));
    }

    /** The directory in which the main-file resides. */
    public void setDirectory(File dir) {
        latexdir = dir;
    }

    /** Sets the verbose flag. */
    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }

    /** Set the BIBTeX flag */
    public void setBibtex(boolean bibtex) {
        this.bibtex = bibtex;
    }

    /** Set the PDFLaTeX flag */
    public void setPdftex(boolean pdftex) {
        this.pdftex = pdftex;
    }

    /** Set the psgztex flag */
    public void setPsgztex(boolean psgztex) {
        this.psgztex = psgztex;
    }

    /** Set the clean flag */
    public void setClean(boolean clean) {
        this.clean = clean;
    }

    public void execute() throws BuildException {
        if (verbose) {
            this.log("Running LaTeX now!");
            this.dump();
        }
        int exitCode = 0;
        log("Calling latex on " + latexfile);
        exitCode = this.runLaTeX();
        log("bibtex...");
        if (bibtex) {
            this.runBIBTeX();
            log("latex");
        }
        this.runLaTeX();

        log("latex");
        this.runLaTeX();

        if (psgztex) {
            log("Creating ps from dvi file...");
            this.runDvips();
            if (pdftex) {
                log("Creating pdf from ps file");
                this.runPs2pdf();
            }
            this.runGzip();
        }
        if (delete == null) {
            delete = this.createDefaultDelete();
        }
        if (verbose) {
            this.log("run Delete");
        }
        delete.execute();
        log("Done.");
    }

    /** Creates the nested delete element */
    public Object createDelete() {
        log("Delete is created");
        delete = (Delete) project.createTask("delete");
        return delete;
    }

    /** Creates the default delete element */
    private Delete createDefaultDelete() {
        Delete delete = null;
        delete = (Delete) project.createTask("delete");
        FileSet fs = new FileSet();
        fs.setDir(this.latexdir);
        for (int i = 0; i < this.deletePatterns.length; i++) {
            fs.createInclude().setName(deletePatterns[i]);
        }
        delete.addFileset(fs);
        delete.setVerbose(true);
        return delete;
    }

    /** Constructs and runs LaTeX-command. */
    private final int runLaTeX() throws BuildException {
        String latex = "latex";
        String[] command = { latex, " &" + latex + " \\nonstopmode\\input{"
                           + latexfile + "}" };
        return this.runCommand(command);
    }

    /** Constructs and runs BibTeX-command. */
    private final int runBIBTeX() throws BuildException {
        String[] command = { "bibtex", bibtexfile };
        return this.runCommand(command);
    }

    /** Constructs and runs dvips-command. */
    private final int runDvips() throws BuildException {
        String[] command = { "dvips", "-D600", "-o", bibtexfile + ".ps", bibtexfile };
        return this.runCommand(command);
    }

    private final int runGzip() throws BuildException {
        String[] command = { "gzip", bibtexfile + ".ps" };
        return this.runCommand(command);
    }

    private final int runPs2pdf() throws BuildException {
        String[] command = { "ps2pdf", bibtexfile + ".ps" };
        return this.runCommand(command);
    }

    /** Runs the command. If verbose is on, then the output of the
        command is logged.
        @param command the command.
     */
    private final int runCommand(String[] command) throws BuildException {
        int exitCode = 0;
        if (verbose) {
            String com = "";
            for (int i = 0; i < command.length; i++) {
                com += command[i];
            }
            log("running command: " + com);
        }
        try {
            Execute exec = null;
            if (verbose) {
                exec = new Execute(new LogStreamHandler(this, Project.MSG_INFO,
                                                        Project.MSG_WARN), null);
            } else {
                exec = new Execute(new PumpStreamHandler(new NirvanaOutputStream()));
            }

            exec.setWorkingDirectory(latexdir);
            exec.setCommandline(command);


            // allow Ant to find the antRun script if necessary
            exec.setAntRun(project);
            exitCode = exec.execute();
        } catch (IOException exp) {
            throw new BuildException(exp);
        }
        if (verbose) {
            log("command exitcode = " + exitCode);
        }
        return exitCode;
    }

    private void cleanIntermediateFiles() {
        String[] suffixes = { "aux", "log", "bbl", "blg", "toc", "lof", "lot" };
    }

    /** Dumps the local attributes. */
    private void dump() {
        this.log("latexfile  = " + latexfile);
        this.log("bibtexfile = " + latexfile);
        this.log("latexdir   = " + latexdir);
        this.log("verbose    = " + verbose);
        this.log("bibtex     = " + bibtex);
        this.log("bibtex     = " + bibtex);
        this.log("clean      = " + clean);
        this.log("delete     = " + delete);
    }

    class NirvanaOutputStream extends OutputStream {
        public void close() throws IOException {
        }

        public void flush() throws IOException {
        }

        public void write(byte[] b, int off, int len) throws IOException {
        }

        public void write(byte[] b) throws IOException {
        }

        public void write(int b) throws IOException {
        }
    }
}