/*
 * Created: 06.05.2004
 *
 */
package de.renew.application;

import de.renew.plugin.PluginProperties;

import de.renew.shadow.ShadowLookup;
import de.renew.shadow.ShadowNet;
import de.renew.shadow.ShadowNetSystem;
import de.renew.shadow.SyntaxException;

import java.io.IOException;
import java.io.ObjectInputStream;

import java.net.URL;

import java.util.Enumeration;
import java.util.Iterator;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;


/**
 * Special Extension that allows a plugin to add its Nets to a running Simulation.
 * Nets are loaded from the resources of the plugin.
 * <p>
 * The feature provided by this extension is experimental!
 * The plugin's nets are inserted into the simulation during the first call
 * to {@link SimulatorPlugin#insertNets}, which leads to a recursive call
 * of that method. As a result, other {@link SimulatorExtension}
 * implementations may observe the wrong order of net insertions.
 * </p>
 * <p>
 * TODO: It would be better if this extension uses the
 * {@link de.renew.net.NetLoader} interface instead of compiling all nets
 * at simulation setup.
 * </p>
 * @author Timo Carl
 * @author (additional comments by Michael Duvigneau)
 **/
public class AddPluginNetsExtension implements SimulatorExtension {
    private JarFile jarRes = null;
    private boolean addNetsAfterNextCompilation = false;

    /**
     * Creates the extension.
     * @param resPath the <code>URL</code> to the <code>.jar</code> file
     *                where the nets reside.
     */
    public AddPluginNetsExtension(URL resPath) {
        if (PluginProperties.debug()) {
            System.out.println("AddPluginNetsExtension with URL " + resPath);
        }
        try {
            jarRes = new JarFile(resPath.getFile());
        } catch (IOException e) {
            System.out.println("Ressource path " + resPath
                               + "is not a valid jar file.");
        }
    }

    public void simulationSetup(SimulationEnvironment env) {
        if (PluginProperties.debug()) {
            System.out.println("AddPluginNetsExtension: Waiting for first net compilation...");
        }
        addNetsAfterNextCompilation = true;
    }

    public void netsCompiled(ShadowNet[] oldNets, ShadowLookup oldLookup) {
        if (addNetsAfterNextCompilation) {
            addNetsAfterNextCompilation = false;
            if (PluginProperties.debug()) {
                System.out.println("AddPluginNetsExtension: Including nets from "
                                   + jarRes);
            }

            //assumes that jarRes is valid because Plugin would not be functional.
            Enumeration entries = jarRes.entries();
            SimulatorPlugin sp = SimulatorPlugin.getCurrent();
            String name = null;
            ShadowNetSystem sns = null;
            while (entries.hasMoreElements()) {
                try {
                    ZipEntry element = (ZipEntry) entries.nextElement();
                    name = element.toString();
                    if (!element.getName().endsWith(".sns")) {
                        continue;
                    }
                    Object o = new ObjectInputStream(jarRes.getInputStream(element))
                               .readObject();
                    if (!(o instanceof ShadowNetSystem)) {
                        continue;
                    }
                    if (PluginProperties.debug()) {
                        System.out.println("adding net " + name);
                    }
                    if (sns == null) {
                        sns = (ShadowNetSystem) o;
                    } else {
                        ShadowNetSystem newSns = (ShadowNetSystem) o;
                        for (Iterator iter = newSns.elements().iterator();
                                     iter.hasNext();) {
                            ShadowNet net = (ShadowNet) iter.next();
                            net.switchNetSystem(sns);
                        }
                    }
                } catch (ClassNotFoundException ce) {
                    System.out.println("This should not happen:" + ce
                                       + "\nAborting.");
                    return;
                } catch (IOException ioe) {
                    System.out.println("Resource" + name
                                       + "is not loadable.\nReason: " + ioe);
                }
            }

            try {
                sp.insertNets(sns);
            } catch (SyntaxException e) {
                System.out.println("Plugin nets contain a syntax error:\n"
                                   + e.getMessage());
                System.out.println("Involved Objects: " + e.errorObjects);
                e.printStackTrace();
            }
        }
    }

    public void simulationTerminated() {
        addNetsAfterNextCompilation = false;
    }
}