package de.renew.application;

import de.renew.engine.simulator.Simulator;

import de.renew.plugin.command.CLCommand;


/**
 * This command line command provides several subcommand to
 * control the simulation engine. This includes single-step mode
 * and consecutive runs as well as simulation termination.
 * <p>
 * </p>
 * SimulationControlCommand.java
 * Created: Wed Jul 16  2003
 *
 * @author Michael Duvigneau
 * @since Renew 2.0
 */
public class SimulationControlCommand implements CLCommand {
    private SimulatorPlugin plugin;

    /**
     * Creates a new <code>SimulationControlCommand</code> instance.
     *
     * @param plugin  neccessary reference to the simulator
     *                plugin instance where this command belongs
     *                to.
     *
     * @throws NullPointerException
     *   if <code>plugin</code> is <code>null</code>.
     **/
    public SimulationControlCommand(SimulatorPlugin plugin) {
        if (plugin == null) {
            throw new NullPointerException("Need SimulatorPlugin reference.");
        }
        this.plugin = plugin;
    }

    /**
     * Controls a simulation or prints a help text.
     * See this class documentation for further details.
     *
     * @param args  one-element-array containing the
     *              simulation control subcommand (see output of
     *              the <code>showSyntax()</code> method).
     **/
    public void execute(String[] args) {
        if ((args == null) || (args.length != 1) || (args[0] == null)) {
            System.err.println("Error: Please give exactly one subcommand. Enter 'help' for a command help.");
        } else {
            handleCommand(args[0]);
        }
    }

    /**
     * Actually interprets the command given to execute.
     **/
    private void handleCommand(String command) {
        if (command.startsWith("help")) {
            showSyntax();
        } else if (command.startsWith("term")) {
            plugin.terminateSimulation();
        } else {
            SimulationEnvironment env = plugin.getCurrentEnvironment();
            if (env == null) {
                System.err.println("Error: There is no current simulation environment.");
            } else {
                Simulator simulator = env.getSimulator();
                if (command.startsWith("run")) {
                    simulator.startRun();
                } else if (command.startsWith("step") || command.equals("")) {
                    simulator.step();
                } else if (command.startsWith("stop")
                                   || command.startsWith("halt")) {
                    simulator.stopRun();
                } else {
                    System.err.println("Error: Unknown command. Enter 'help' for a command help.");
                }
            }
        }
    }

    /* Non-JavaDoc: Specified by the CLCommand interface. */
    public String getDescription() {
        return "control the simulation by subcommands (e.g. step, run, stop)";
    }

    /**
     * Prints command-line help for this command to <code>System.out</code>.
     **/
    public void showSyntax() {
        System.out.println("Simulation control commands:");
        System.out.println("help   Displays this help.");
        System.out.println("run    Runs the simulation continuously.");
        System.out.println("step   Executes only one simulation step.");
        System.out.println("stop   Stops the simulation, but it remains initialized.");
        System.out.println("halt   Same as stop.");
        System.out.println("term   Terminates the simulation.");
    }
}