package de.renew.application;

import de.renew.database.SetupHelper.SimulationState;

import de.renew.engine.searcher.SilentTracerFactory;
import de.renew.engine.searcher.TracerFactory;

import de.renew.net.NetInstance;

import de.renew.plugin.command.CLCommand;

import de.renew.shadow.ShadowNetSystem;

import de.renew.util.ParameteredCommandLine;

import java.io.FileInputStream;
import java.io.ObjectInput;
import java.io.ObjectInputStream;


/**
 * This command line command sets up a simulation based on the
 * given shadow net system file and primary net name. It can also
 * print a usage help text.
 * <p>
 * This command always tries to reestablish a database backed state
 * if the required properties are set. If such a state is found,
 * the primary net does not get instantiated (as it would be its
 * second instantiation).
 * </p>
 * StartSimulationCommand.java
 * Created: Wed Jul 16  2003
 *
 * @author Michael Duvigneau
 * @since Renew 2.0
 **/
public class StartSimulationCommand implements CLCommand {
    private SimulatorPlugin plugin;

    /**
     * Creates a new <code>StartSimulationCommand</code> instance.
     *
     * @param plugin  neccessary reference to the simulator
     *                plugin instance where this command belongs
     *                to.
     *
     * @throws NullPointerException
     *   if <code>plugin</code> is <code>null</code>.
     **/
    public StartSimulationCommand(SimulatorPlugin plugin) {
        if (plugin == null) {
            throw new NullPointerException("Need SimulatorPlugin reference.");
        }
        this.plugin = plugin;
    }

    /**
     * Sets up a simulation or prints a help text.
     * See this class documentation for further details.
     *
     * @param args  arguments to this command, specifying the sns
     *              file, the name of the main net and other
     *              optional arguments (see output of the
     *              <code>showSyntax()</code> method).
     **/
    public void execute(String[] args) {
        // This flag tells whether we have to terminate the
        // simulation if an error occurs.
        boolean mayBeInconsistent = false;
        try {
            ParameteredCommandLine line = new ParameteredCommandLine(args,
                                                                     new String[] { "-h", "-i" },
                                                                     new int[] { 0, 0 });
            String[] basicArgs = line.getRemainingArgs();

            if (basicArgs.length != 2 || line.hasParameter("-h")) {
                showSyntax();
            } else {
                String netSystemFileName = basicArgs[0];
                String primaryNetName = basicArgs[1];

                // Load the shadow net system
                FileInputStream stream = new FileInputStream(netSystemFileName);
                ObjectInput input = new ObjectInputStream(stream);
                ShadowNetSystem netSystem;
                try {
                    netSystem = (ShadowNetSystem) input.readObject();
                } catch (ClassCastException e) {
                    throw new IllegalArgumentException("Invalid shadow net system (in file "
                                                       + netSystemFileName
                                                       + "): " + e);
                } catch (ClassNotFoundException e) {
                    throw new IllegalArgumentException("Invalid shadow net system (in file "
                                                       + netSystemFileName
                                                       + "): " + e);
                }

                // Set up the simulation.
                TracerFactory tracerFactory = new SilentTracerFactory();
                plugin.setupSimulation(null, tracerFactory);
                mayBeInconsistent = true;
                plugin.insertNets(netSystem);

                // Create or restore net instances (depending on database).
                SimulationState state = plugin.restoreStateFromDatabase();
                if (!state.wasSimulationInited()) {
                    NetInstance primaryInstance = plugin.createNetInstance(primaryNetName);
                    System.out.println("Simulation set up, created net instance "
                                       + primaryInstance + ".");
                } else {
                    System.out.println("Simulation set up, restored state from database.");
                }

                // Run the simulation, if requested.
                if (!line.hasParameter("-i")) {
                    SimulationEnvironment env = plugin.getCurrentEnvironment();
                    env.getSimulator().startRun();
                    System.out.println("Simulation running.");
                }
            }
        } catch (Exception e) {
            System.err.println("Could not start simulation: " + e);
            e.printStackTrace();
            if (mayBeInconsistent) {
                System.err.println("Cleaning up.");
                plugin.terminateSimulation();
            }
        }
    }

    /* Non-JavaDoc: Specified by the CLCommand interface. */
    public String getDescription() {
        return "set up a simulation with given nets (-h for help).";
    }

    /**
     * Prints command-line help for this command to <code>System.out</code>.
     **/
    public void showSyntax() {
        System.out.println("Parameters: <net system> <primary net> [-i]");
        System.out.println("  <net system>  : The file name of the exported shadow net system to load.");
        System.out.println("  <primary net> : The name of the primary net to create an instance of.");
        System.out.println("  -i            : Initialize the simulation only, don't run it.");
        System.out.println("To configure the simulation engine and installed extensions, use the");
        System.out.println("available plugin properties before starting the simulation (e.g. ");
        System.out.println(SimulatorPlugin.MODE_PROP_NAME + " or "
                           + SimulatorPlugin.EAGER_PROP_NAME + ").");
    }
}