package de.renew.remote;

import de.renew.application.SimulationEnvironment;

import de.renew.engine.searcher.TriggerCollection;
import de.renew.engine.searcher.Triggerable;
import de.renew.engine.simulator.Binding;
import de.renew.engine.simulator.SimulatorHelper;

import de.renew.net.NetElementID;
import de.renew.net.TransitionInstance;
import de.renew.net.event.FiringEvent;
import de.renew.net.event.TransitionEventListener;

import java.rmi.RemoteException;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;


/**
 * This class implements the <code>TransitionInstanceAccessor</code>
 * interface and nothing more than needed to implement it.
 * <p>
 * </p>
 * TransitionInstanceAccessorImpl.java
 * Created: Mon Jul 17  2000
 * @author Michael Duvigneau
 */
public class TransitionInstanceAccessorImpl extends ObjectAccessorImpl
        implements TransitionInstanceAccessor, TransitionEventListener,
                           Triggerable {

    /**
     * The set of registered remote event listeners that want
     * to be informed about transition events.
     */
    private Set firingListeners = Collections.synchronizedSet(new HashSet());

    /**
     * The set of registered triggerable forwarders that want
     * to be informed about search proposals.
     */
    private Set proposalListeners = Collections.synchronizedSet(new HashSet());

    /**
     * The triggers of this triggerable.
     */
    private TriggerCollection triggers = new TriggerCollection(this);

    /**
     * Creates a new transition instance accessor for the given transition
     * instance.
     *
     * @param transitionInstance the transition instance to access
     * @exception java.rmi.RemoteException if an RMI failure occured.
     */
    public TransitionInstanceAccessorImpl(TransitionInstance transitionInstance,
                                          SimulationEnvironment environment)
            throws RemoteException {
        super(transitionInstance, environment);
    }

    /**
     * Registers the given remote event listener so that it will
     * receive future remote event messages concerning the accessed
     * transition instance.
     * <p>
     * If the listener is registered already, the additional
     * registration try is ignored. The <code>equals()</code>
     * method serves as indicator to allow the mechanism to
     * work also for remote objects.
     * </p>
     * <p>
     * This method is specified by the <code>RemoteEventProducer</code>
     * interface which is required by the <code>TransitionInstanceAccessor</code>
     * interface.
     * </p>
     *
     * @param listener the listener to register
     * @exception java.rmi.RemoteException if a RMI failure occured.
     */
    public void addRemoteEventListener(RemoteEventListener listener)
            throws RemoteException {
        synchronized (firingListeners) {
            // We register this accessor as TransitionEventListener
            // when adding the first remote event listener.
            if (firingListeners.isEmpty()) {
                ((TransitionInstance) object).addTransitionEventListener(this);
            }

            firingListeners.add(listener);
        }
    }

    /**
     * {@inheritDoc}
     **/
    public BindingAccessor[] findAllBindings(TriggerableForwarder triggerableForwarder)
            throws RemoteException {
        synchronized (proposalListeners) {
            proposalListeners.add(triggerableForwarder);
        }

        BindingAccessor[] bindings;
        try {
            triggers.clear();
            Collection allBindings = SimulatorHelper.findAllBindings((TransitionInstance) object,
                                                                     this);
            Iterator iter = allBindings.iterator();

            bindings = new BindingAccessor[allBindings.size()];
            for (int i = 0; iter.hasNext(); i++) {
                bindings[i] = new BindingAccessorImpl((Binding) iter.next());
            }
        } catch (RemoteException e) {
            e.printStackTrace(System.err);
            bindings = new BindingAccessor[0];
        }

        return bindings;
    }

    /**
     * {@inheritDoc}
     **/
    public void forgetBindings(TriggerableForwarder triggerableForwarder)
            throws RemoteException {
        synchronized (proposalListeners) {
            proposalListeners.remove(triggerableForwarder);
            if (proposalListeners.isEmpty()) {
                triggers.clear();
            }
        }
    }

    /**
     * {@inheritDoc}
     **/
    public boolean fireOneBinding(TracerAccessor tracer)
            throws RemoteException {
        boolean asynchronous = !getEnvironment().getSimulator().isSequential();
        return ((TransitionInstance) object).fireOneBinding(new RemoteTracer(tracer),
                                                            asynchronous);
    }

    /**
     * The common reaction to all transition events is to send a
     * remote event to all currently registered remote event
     * listeners.
     */
    private void fireRemoteEvent() {
        Iterator iterator;
        RemoteEventListener listener;
        synchronized (firingListeners) {
            iterator = firingListeners.iterator();
            while (iterator.hasNext()) {
                listener = (RemoteEventListener) iterator.next();
                try {
                    listener.update();
                } catch (RemoteException e) {
                    System.err.println("TransitionInstanceAccessor: Remote event to "
                                       + listener
                                       + " probably got lost due to " + e);
                }
            }
        }
    }

    /**
     * Converts the local transition event into a remote event and
     * forwards it to all remote listeners.
     */
    public void firingComplete(FiringEvent event) {
        fireRemoteEvent();
    }

    /**
     * Converts the local transition event into a remote event and
     * forwards it to all remote listeners.
     */
    public void firingStarted(FiringEvent event) {
        fireRemoteEvent();
    }

    /* This method is specified by the TransitionInstanceAccessor interface. */
    public NetElementID getID() throws RemoteException {
        return ((TransitionInstance) object).getTransition().getID();
    }

    /* This method is specified by the TransitionInstanceAccessor interface. */
    public TransitionAccessor getTransition() throws RemoteException {
        return new TransitionAccessorImpl(((TransitionInstance) object)
                                          .getTransition(), getEnvironment());
    }

    /**
     * Returns the transition instance, if the caller knows that
     * this is the local representation. This is required for setting breakpoints.
     * @return The transition instance.
     */
    public TransitionInstance getTransitionInstance() {
        return (TransitionInstance) object;
    }

    /* This method is specified by the TransitionInstanceAccessor interface. */
    public NetInstanceAccessor getNetInstance() throws RemoteException {
        return new NetInstanceAccessorImpl(((TransitionInstance) object)
                                           .getNetInstance(), getEnvironment());
    }

    /**
     * Returns whether the transition instance is currently firing
     * at least one occurrance.
     * @return Whether it is firing.
     * @exception java.rmi.RemoteException if an RMI failure occured.
     */
    public boolean isFiring() throws RemoteException {
        return ((TransitionInstance) object).isFiring();
    }

    /**
     * Called when a trigger proposes a new search.
     */
    public synchronized void proposeSearch() {
        List triggerableForwarders;
        synchronized (proposalListeners) {
            triggerableForwarders = new ArrayList(proposalListeners);
            proposalListeners.clear();
            triggers.clear();
        }
        for (Iterator i = triggerableForwarders.iterator(); i.hasNext();) {
            TriggerableForwarder forwarder = null;
            try {
                forwarder = ((TriggerableForwarder) i.next());
                forwarder.proposeSearch();
            } catch (RemoteException e) {
                System.err.println("TransitionInstanceAccessor: Search proposal to "
                                   + forwarder + " probably got lost due to "
                                   + e);
            }
        }
    }

    /**
     * Unregisters the given remote event listener so that it will
     * not receive future remote event messages from this transition
     * instance accessor.
     * <p>
     * All listeners that equal the specified one will be
     * unregistered. If the listener was not registered, the
     * unregistration try is ignored.
     * </p>
     * <p>
     * This method is specified by the <code>RemoteEventProducer</code>
     * interface which is required by the <code>TransitionInstanceAccessor</code>
     * interface.
     * </p>
     *
     * @param listener the listener to unregister
     * @exception java.rmi.RemoteException if a RMI failure occured.
     */
    public void removeRemoteEventListener(RemoteEventListener listener)
            throws RemoteException {
        synchronized (firingListeners) {
            firingListeners.remove(listener);


            // We unregister this accessor as TransitionEventListener
            // when removing the last remote event listener.
            if (firingListeners.isEmpty()) {
                ((TransitionInstance) object).removeTransitionEventListener(this);
            }
        }
    }

    /**
     * Returns the triggers of this triggerable.
     * @return The triggers of this triggerable.
     */
    public TriggerCollection triggers() {
        return triggers;
    }

    /**
     * This class does not use synchronous notifications, because
     * it might lead to deadlocks for some applications. Asynchronous
     * notifications are always safe, although they might arrive slightly
     * later.
     *
     * @return a <code>boolean</code> value
     */
    public boolean wantSynchronousNotification() {
        return false;
    }
}