package de.renew.shadow;

import de.renew.net.Net;
import de.renew.net.NetLoader;
import de.renew.net.NetNotFoundException;


/**
 * A generic net loader implementation which delegates to a
 * {@link ShadowNetLoader}. The {@link ShadowNetSystem} given on
 * creation of this net loader has to know about its compiler.
 * The net system will be kept up to date whenever a new net is
 * loaded through this net loader.
 * <p>
 * </p>
 * DefaultCompiledNetLoader.java
 * Created: Tue Dec  5 2001
 * @author Michael Duvigneau
 **/
public class DefaultCompiledNetLoader implements NetLoader {
    private ShadowNetSystem netSystem;
    private ShadowNetLoader netLoader;

    /**
     * Creates a default compiled net loader for the given shadow
     * net system, delegating to the given shadow net loader.
     *
     * @exception NullPointerException
     *  if either <code>netSystem</code> or <code>netLoader</code>
     *  is <code>null</code>.
     **/
    public DefaultCompiledNetLoader(ShadowNetSystem netSystem,
                                    ShadowNetLoader netLoader) {
        if (netSystem == null) {
            throw new NullPointerException("Missing ShadowNetSystem.");
        } else if (netLoader == null) {
            throw new NullPointerException("Missing ShadowNetLoader.");
        }
        this.netSystem = netSystem;
        this.netLoader = netLoader;
    }

    /**
     * Creates a default compiled net loader for the given shadow
     * net system. The reference to a shadow net loader for
     * delegation is extracted from the net system immediately.
     *
     * @exception NullPointerException
     *  if the <code>netSystem</code> is <code>null</code> or does
     *  not known about a shadow net loader.
     **/
    public DefaultCompiledNetLoader(ShadowNetSystem netSystem) {
        if (netSystem == null) {
            throw new NullPointerException("Missing ShadowNetSystem.");
        }
        this.netSystem = netSystem;
        this.netLoader = netSystem.getNetLoader();
        if (netLoader == null) {
            throw new NullPointerException("Missing ShadowNetLoader.");
        }
    }

    public Net loadNet(String netName) throws NetNotFoundException {
        try {
            ShadowNet shadowNet = netLoader.loadShadowNet(netName, netSystem);
            ShadowLookup lookup = netSystem.compileMore();
            if (lookup != null) {
                Net result = lookup.getNet(netName);
                if (result != null) {
                    return result;
                }
            }
        } catch (NetNotFoundException e) {
            // Just forward the information from the shadow net loader...
            throw e;
        } catch (Exception e) {
            throw new NetNotFoundException(netName, e);
        }
        throw new NetNotFoundException(netName);
    }
}